
//
//  ID3v2 ^O擾
//
//  Written by Otachan
//  http://www.aikis.or.jp/~otachan/
//

#define	STRICT

#include <windows.h>
#include <mbctype.h>

#include "CommonFunc.h"
#include "ID3v2Info.h"

#define	TITLEBUFLEN		(MAX_PATH - 10)
#define	MAX_MUSICTEXT	256

bool
GetTitleFromID3v2(char* FileName, char* RetTitle, const char* format)
{
	::SetLastError(NO_ERROR);

	HANDLE	hF = ::CreateFile(FileName, GENERIC_READ, FILE_SHARE_READ,
								NULL, OPEN_EXISTING, FILE_FLAG_RANDOM_ACCESS, NULL);

	if(::GetLastError() != NO_ERROR) return false;

	bool	RetCode;
	char	Title[MAX_MUSICTEXT];
	char	Author[MAX_MUSICTEXT];
	char	Comment[MAX_MUSICTEXT];
	char	Album[MAX_MUSICTEXT];
	char	Year[MAX_MUSICTEXT];
	char	Genre[MAX_MUSICTEXT];
	char	TrackNo[MAX_MUSICTEXT];
	char	Compo[MAX_MUSICTEXT];
	char	OrgArt[MAX_MUSICTEXT];
	char	Copyright[MAX_MUSICTEXT];
	char	Encoder[MAX_MUSICTEXT];
	DWORD	FileSize = ::GetFileSize(hF, NULL);

	if(FileSize >= 10) {
		unsigned char	TAG[10];
		DWORD	ReadByte;

		memset(TAG, '\0', 3);
		::ReadFile(hF, TAG, 10, &ReadByte, NULL);

		unsigned char	TagVersion;

		if((*TAG == 'I') && (*(TAG + 1) == 'D') && (*(TAG + 2) == '3') &&
												 ((TagVersion = *(TAG + 3)) >= 3)) {
			UINT	TagSize =	(*(TAG + 6) << 21) +
								(*(TAG + 7) << 14) +
								(*(TAG + 8) << 7) +
								*(TAG + 9);

			if(FileSize >= (10 + TagSize)) {
				const unsigned char	TagFlag = *(TAG + 5);
				const bool	TagUnsynchronization = (TagFlag & 0x80) != 0;
				const bool	TagExtendedHeader = (TagFlag & 0x40) != 0;
				const bool	TagFooter = (TagFlag & 0x10) != 0;
				unsigned char*	TagData = new unsigned char[TagSize];

				::ReadFile(hF, TagData, TagSize, &ReadByte, NULL);
				::CloseHandle(hF);

				if(TagUnsynchronization) {
					TagSize = ID3v2DecodeUnsynchronization(TagData, TagSize, true);
				}

				unsigned char*	TagDataPnt = TagData;
				unsigned char*	MaxTagDataPnt = TagDataPnt + TagSize;

				if(TagExtendedHeader) {
					MaxTagDataPnt -=	(*(TagDataPnt + 6) << 24) +
										(*(TagDataPnt + 7) << 16) +
										(*(TagDataPnt + 8) << 8) +
										*(TagDataPnt + 9);
					TagDataPnt +=	(*(TagDataPnt + 0) << 24) +
									(*(TagDataPnt + 1) << 16) +
									(*(TagDataPnt + 2) << 8) +
									*(TagDataPnt + 3) +
									4;
				}

				if(TagFooter) MaxTagDataPnt -= 10;

				bool	TitleSc = false;
				bool	AuthorSc = false;
				bool	CommentSc = false;
				bool	AlbumSc = false;
				bool	YearSc = false;
				bool	GenreSc = false;
				bool	TrackNoSc = false;
				bool	CompoSc = false;
				bool	OrgArtSc = false;
				bool	CopyrightSc = false;
				bool	EncoderSc = false;

				while((TagDataPnt < MaxTagDataPnt) &&
										((TitleSc == false) ||
										 (AuthorSc == false) ||
										 (CommentSc == false) ||
										 (AlbumSc == false) ||
										 (YearSc == false) ||
										 (GenreSc == false) ||
										 (TrackNoSc == false) ||
										 (CompoSc == false) ||
										 (OrgArtSc == false) ||
										 (CopyrightSc == false) ||
										 (EncoderSc == false))) {
					const unsigned char FrameID0 = *TagDataPnt++;

					if(FrameID0 == '\0') break;

					const unsigned char FrameID1 = *TagDataPnt++;
					const unsigned char FrameID2 = *TagDataPnt++;
					const unsigned char FrameID3 = *TagDataPnt++;

					UINT	FrameSize =	(*(TagDataPnt + 0) << 24) +
										(*(TagDataPnt + 1) << 16) +
										(*(TagDataPnt + 2) << 8) +
										*(TagDataPnt + 3);
					const UINT	AddFrameSize = FrameSize;
					const unsigned char	FrameFlag2 = *(TagDataPnt + 5);
					bool	FrameCompression;
					bool	FrameCoded;
					bool	FrameUnsynchronization;

					if(TagVersion >= 4) {	// Ver. 2.4
						FrameCompression = (FrameFlag2 & 0x08) != 0;
						FrameCoded = (FrameFlag2 & 0x04) != 0;
						FrameUnsynchronization = (FrameFlag2 & 0x02) != 0;
					} else {				// Ver. 2.3
						FrameCompression = (FrameFlag2 & 0x80) != 0;
						FrameCoded = (FrameFlag2 & 0x40) != 0;
						FrameUnsynchronization = false;
					}

					TagDataPnt += 6;

					if((FrameCompression == false) && (FrameCoded == false)) {
						if(FrameID0 == 'T') {
							char*	Buff;
							UINT	BuffSize;

							if((TitleSc == false) &&			(FrameID1 == 'I') &&
																(FrameID2 == 'T') &&
																(FrameID3 == '2')) {
								Buff = Title;
								BuffSize = MAX_MUSICTEXT;
								TitleSc = true;
							} else if((AuthorSc == false) &&	(FrameID1 == 'P') &&
																(FrameID2 == 'E') &&
																(FrameID3 == '1')) {
								Buff = Author;
								BuffSize = MAX_MUSICTEXT;
								AuthorSc = true;
							} else if((AlbumSc == false) &&		(FrameID1 == 'A') &&
																(FrameID2 == 'L') &&
																(FrameID3 == 'B')) {
								Buff = Album;
								BuffSize = MAX_MUSICTEXT;
								AlbumSc = true;
							} else if((YearSc == false) &&		(FrameID1 == 'Y') &&
																(FrameID2 == 'E') &&
																(FrameID3 == 'R')) {
								Buff = Year;
								BuffSize = MAX_MUSICTEXT;
								YearSc = true;
							} else if((GenreSc == false) &&		(FrameID1 == 'C') &&
																(FrameID2 == 'O') &&
																(FrameID3 == 'N')) {
								Buff = Genre;
								BuffSize = MAX_MUSICTEXT;
								GenreSc = true;
							} else if((TrackNoSc == false) &&	(FrameID1 == 'R') &&
																(FrameID2 == 'C') &&
																(FrameID3 == 'K')) {
								Buff = TrackNo;
								BuffSize = MAX_MUSICTEXT;
								TrackNoSc = true;
							} else if((CompoSc == false) &&		(FrameID1 == 'C') &&
																(FrameID2 == 'O') &&
																(FrameID3 == 'M')) {
								Buff = Compo;
								BuffSize = MAX_MUSICTEXT;
								CompoSc = true;
							} else if((OrgArtSc == false) &&	(FrameID1 == 'O') &&
																(FrameID2 == 'P') &&
																(FrameID3 == 'E')) {
								Buff = OrgArt;
								BuffSize = MAX_MUSICTEXT;
								OrgArtSc = true;
							} else if((CopyrightSc == false) &&	(FrameID1 == 'C') &&
																(FrameID2 == 'O') &&
																(FrameID3 == 'P')) {
								Buff = Copyright;
								BuffSize = MAX_MUSICTEXT;
								CopyrightSc = true;
							} else if((EncoderSc == false) &&	(FrameID1 == 'E') &&
																(FrameID2 == 'N') &&
																(FrameID3 == 'C')) {
								Buff = Encoder;
								BuffSize = MAX_MUSICTEXT;
								EncoderSc = true;
							} else {
								BuffSize = 0;
							}

							if(BuffSize) {
								ID3v2FrameTEXT(TagDataPnt, FrameSize, Buff, BuffSize,
															FrameUnsynchronization);
							}
						} else if((CommentSc == false) &&	(FrameID0 == 'C') &&
															(FrameID1 == 'O') &&
															(FrameID2 == 'M') &&
															(FrameID3 == 'M')) {
							ID3v2FrameCOMM(TagDataPnt, FrameSize, Comment,
														FrameUnsynchronization);
							CommentSc = true;
						}
					}
					TagDataPnt += AddFrameSize;
				}

				delete[] TagData;

				if(TitleSc == false) *Title = '\0';
				if(AuthorSc == false) *Author = '\0';
				if(CommentSc == false) *Comment = '\0';
				if(AlbumSc == false) *Album = '\0';
				if(YearSc == false) *Year = '\0';
				if(GenreSc == false) *Genre = '\0';
				if(TrackNoSc == false) *TrackNo = '\0';
				if(CompoSc == false) *Compo = '\0';
				if(OrgArtSc == false) *OrgArt = '\0';
				if(CopyrightSc == false) *Copyright = '\0';
				if(EncoderSc == false) *Encoder = '\0';

				RetCode = true;
			} else {
				RetCode = false;
			}
		} else {
			RetCode = false;
		}
	} else {
		RetCode = false;
	}

	if(RetCode == false) {
		::CloseHandle(hF);
		return false;
	}

	const char*	p;
	char*	q = RetTitle;;
	int		remain = TITLEBUFLEN;
	int		StrLen;

	for(p = format; *p != '\0' && remain > 0;) {
		if(*p != '%') {
			*q++ = *p++;
			remain--;
			continue;
		}

		switch(*++p)
		{
		case '%':
			*q++ = '%';
			remain--;
			break;
		case '1':
			Strncpy(q, Author, remain);
			StrLen = strlen(Author);
			remain -= StrLen;
			q += StrLen;
			break;
		case '2':
			Strncpy(q, Title, remain);
			StrLen = strlen(Title);
			remain -= StrLen;
			q += StrLen;
			break;
		case '3':
			Strncpy(q, Album, remain);
			StrLen = strlen(Album);
			remain -= StrLen;
			q += StrLen;
			break;
		case '4':
			Strncpy(q, Year, remain);
			StrLen = strlen(Year);
			remain -= StrLen;
			q += StrLen;
			break;
		case '5':
			Strncpy(q, Comment, remain);
			StrLen = strlen(Comment);
			remain -= StrLen;
			q += StrLen;
			break;
		case '6':
			{
				char*	GenreBuff = Genre;

				if(*GenreBuff == '(') {
					for(; *GenreBuff; GenreBuff++) {
						if(*GenreBuff == ')') {
							GenreBuff++;
							break;
						}
					}
				}
				Strncpy(q, GenreBuff, remain);
				StrLen = strlen(GenreBuff);
				remain -= StrLen;
				q += StrLen;
			}
			break;
		case '7':
			{
				char	Name[MAX_PATH];
				char	NameOnly[MAX_PATH];

				CutPathFileName(FileName, NULL, Name);
				CutFileNameExt(Name, NameOnly, NULL);
				Strncpy(q, NameOnly, remain);
				StrLen = strlen(NameOnly);
				remain -= StrLen;
				q += StrLen;
			}
			break;
		case '8':
			{
				char	Path[MAX_PATH];

				CutPathFileName(FileName, Path, NULL);
				Strncpy(q, Path, remain);
				StrLen = strlen(Path);
				remain -= StrLen;
				q += StrLen;
			}
			break;
		case '9':
			{
				char	Ext[MAX_PATH];

				CutFileNameExt(FileName, NULL, Ext);
				Strncpy(q, Ext, remain);
				StrLen = strlen(Ext);
				remain -= StrLen;
				q += StrLen;
			}
			break;
		case 'a':
			Strncpy(q, TrackNo, remain);
			StrLen = strlen(TrackNo);
			remain -= StrLen;
			q += StrLen;
			break;
		case 'b':
			Strncpy(q, Compo, remain);
			StrLen = strlen(Compo);
			remain -= StrLen;
			q += StrLen;
			break;
		case 'c':
			Strncpy(q, OrgArt, remain);
			StrLen = strlen(OrgArt);
			remain -= StrLen;
			q += StrLen;
			break;
		case 'd':
			Strncpy(q, Copyright, remain);
			StrLen = strlen(Copyright);
			remain -= StrLen;
			q += StrLen;
			break;
		case 'e':
			Strncpy(q, Encoder, remain);
			StrLen = strlen(Encoder);
			remain -= StrLen;
			q += StrLen;
			break;
		default:
			if (remain >= 1) { *q++ = '%'; remain--; }
			if (remain >= 1) { *q++ = *p;  remain--; }
			break;
		}

		p++;
	}

	*q = '\0';

	return true;
}

inline void
ID3v2FrameTEXT(unsigned char* TagDataPnt, UINT FrameSize, char* Buff,
							const UINT BuffSize, const bool FrameUnsynchronization)
{
	if(FrameUnsynchronization) {
		FrameSize = ID3v2DecodeUnsynchronization(TagDataPnt, FrameSize, false);
	}

	const UINT	FrameStrSize = FrameSize - 1;
	char*	FrameStrPnt = reinterpret_cast<char*>(TagDataPnt + 1);
	const UINT	StrCode = *TagDataPnt;
	bool	UnicodeUseBOM;

	switch(StrCode) {
	case 0x00:		// ISO-8859-1
		KanjiStrncpy(Buff, FrameStrPnt, Min(FrameStrSize, BuffSize - 1));
		break;
	case 0x01:		// Unicode(UTF-16 Use BOM)
		UnicodeUseBOM = true;
	case 0x02:		// Unicode(UTF-16 BigEndian)
		if(StrCode == 0x02) UnicodeUseBOM = false;

		WideToMultiStrcpy(Buff, reinterpret_cast<WCHAR*>(FrameStrPnt),
							BuffSize, FrameStrSize, UnicodeUseBOM, true);
		break;
	case 0x03:		// Unicode(UTF-8)
		UTF8ToMultiStrcpy(Buff, reinterpret_cast<unsigned char*>(FrameStrPnt),
														BuffSize, FrameStrSize);
		break;
	}
}

inline void
ID3v2FrameCOMM(unsigned char* TagDataPnt, UINT FrameSize, char* BuffComment,
												const bool FrameUnsynchronization)
{
	if(FrameUnsynchronization) {
		FrameSize = ID3v2DecodeUnsynchronization(TagDataPnt, FrameSize, false);
	}

	const int	FrameStrSize = FrameSize - 1 - 3;
	char*	FrameStrPnt = reinterpret_cast<char*>(TagDataPnt + 1 + 3);
	const UINT	StrCode = *TagDataPnt;
	int		SimpleStrSize;
	int		DetailStrSize;
	bool	UnicodeUseBOM;

	switch(StrCode) {
	case 0x00:		// ISO-8859-1
		SimpleStrSize = strlen(FrameStrPnt);
		DetailStrSize = FrameStrSize - SimpleStrSize - 1;

		if(DetailStrSize >= 1) {
			KanjiStrncpy(BuffComment, FrameStrPnt + SimpleStrSize + 1,
									Min(DetailStrSize, MAX_MUSICTEXT - 1));
		} else {
			KanjiStrncpy(BuffComment, FrameStrPnt,
									Min(SimpleStrSize, MAX_MUSICTEXT - 1));
		}
		break;
	case 0x01:		// Unicode(UTF-16 Use BOM)
		UnicodeUseBOM = true;
	case 0x02:		// Unicode(UTF-16 BigEndian)
		if(StrCode == 0x02) UnicodeUseBOM = false;

		SimpleStrSize = WideStrSize(reinterpret_cast<WCHAR*>(FrameStrPnt),
																FrameStrSize);
		DetailStrSize = FrameStrSize - SimpleStrSize;

		if(DetailStrSize >= 2) {
			WideToMultiStrcpy(BuffComment,
					reinterpret_cast<WCHAR*>(FrameStrPnt + SimpleStrSize),
					MAX_MUSICTEXT, DetailStrSize, UnicodeUseBOM, true);
		} else {
			WideToMultiStrcpy(BuffComment,
					reinterpret_cast<WCHAR*>(FrameStrPnt),
					MAX_MUSICTEXT, SimpleStrSize, UnicodeUseBOM, true);
		}
		break;
	case 0x03:		// Unicode(UTF-8)
		SimpleStrSize = strlen(FrameStrPnt) + 1;
		DetailStrSize = FrameStrSize - SimpleStrSize;

		if(DetailStrSize >= 1) {
			UTF8ToMultiStrcpy(BuffComment,
					reinterpret_cast<unsigned char*>(FrameStrPnt + SimpleStrSize),
					MAX_MUSICTEXT, DetailStrSize);
		} else {
			UTF8ToMultiStrcpy(BuffComment,
					reinterpret_cast<unsigned char*>(FrameStrPnt),
					MAX_MUSICTEXT, SimpleStrSize);
		}
		break;
	}
}

UINT
ID3v2DecodeUnsynchronization(unsigned char* Data, const UINT Size, const bool SetZero)
{
	unsigned char*	WriteData = Data;
	UINT	DecodeSize = 0;
	bool	FullBits = false;

	for(UINT Idx = 0; Idx < Size; Idx++) {
		const unsigned char	DataPnt = *Data++;

		if(DataPnt == 0xff) {
			FullBits = true;
		} else {
		 	if(FullBits) {
				FullBits = false;
				if(DataPnt == 0x00) continue;
			}
		}

		*WriteData++ = DataPnt;
		DecodeSize++;
	}

	if(SetZero) memset(WriteData, 0, Size - DecodeSize);

	return DecodeSize;
}

