/* 
 * Mpeg Layer-1,2,3 audio decoder 
 * ------------------------------
 * copyright (c) 1995,1996,1997 by Michael Hipp, All rights reserved.
 * See also 'README'
 *
 * slighlty optimized for machines without autoincrement/decrement.
 * The performance is highly compiler dependend. Maybe
 * the decode.c version for 'normal' processor may be faster
 * even for Intel processors.
 */

#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "mpg123.h"
#include "mpglib.h"

__inline int
WRITE_SAMPLE_INT32(int* samples, real sum)
{
	int		clip;
//	__int64	v = (__int64)sum;
	__int64	v;

	_asm {
		fld   qword ptr [sum]
		fistp qword ptr [v]
	}

/*
	double	dtemp = ((((65536.0 * 65536.0 * 16.0)+
						(65536.0 * 65536.0 * 16.0 * 0.5)) * 65536.0)) + sum;
	__int64	v = (__int64)(*((unsigned __int64*)&dtemp) & 0xfffffffffffffUi64) -
																	0x8000000000000i64;
*/

	if(v > 0x7fffffffi64) {
		v = 0x7fffffffi64;
		clip = 1;
	} else if(v < -0x80000000i64) {
		v = -0x80000000i64;
		clip = 1;
	} else {
		clip = 0;
	}

	*samples = (int)v;

	return clip;
}

int synth_1to1_mono_int32(struct mpstr *mp,real *bandPtr,unsigned char *samples,int *pnt)
{
  int samples_tmp[64];
  int* tmp1 = samples_tmp;
  int i,ret;
  int pnt1 = 0;

  ret = synth_1to1_int32(mp,bandPtr,0,(unsigned char*)samples_tmp,&pnt1);
  samples += *pnt;

  for(i=0;i<32;i++) {
    *((int*)samples) = *tmp1;
    samples += 4;
    tmp1 += 2;
  }

  (*pnt) += 128;

  return ret;
}

int synth_1to1_int32(struct mpstr *mp,real *bandPtr,int channel,unsigned char *out,int *pnt)
{
  static const int step = 2;
  int bo;
  int *samples = (int*)(out + *pnt);

  real *b0,(*buf)[0x110];
  int clip = 0; 
  int bo1;

  bo = mp->synth_bo;

  if(!channel) {
    bo--;
    bo &= 0xf;
    buf = mp->synth_buffs[0];
  }
  else {
    samples++;
    buf = mp->synth_buffs[1];
  }

  if(bo & 0x1) {
    b0 = buf[0];
    bo1 = bo;
    dct64(buf[1]+((bo+1)&0xf),buf[0]+bo,bandPtr);
  }
  else {
    b0 = buf[1];
    bo1 = bo+1;
    dct64(buf[0]+bo,buf[1]+bo+1,bandPtr);
  }

  mp->synth_bo = bo;

  {
    register int j;
    real *window = decwin + 16 - bo1;

    for (j=16;j;j--,b0+=0x10,window+=0x20,samples+=step)
    {
      real sum;
      sum  = window[0x0] * b0[0x0];
      sum -= window[0x1] * b0[0x1];
      sum += window[0x2] * b0[0x2];
      sum -= window[0x3] * b0[0x3];
      sum += window[0x4] * b0[0x4];
      sum -= window[0x5] * b0[0x5];
      sum += window[0x6] * b0[0x6];
      sum -= window[0x7] * b0[0x7];
      sum += window[0x8] * b0[0x8];
      sum -= window[0x9] * b0[0x9];
      sum += window[0xA] * b0[0xA];
      sum -= window[0xB] * b0[0xB];
      sum += window[0xC] * b0[0xC];
      sum -= window[0xD] * b0[0xD];
      sum += window[0xE] * b0[0xE];
      sum -= window[0xF] * b0[0xF];

      clip += WRITE_SAMPLE_INT32(samples,sum);
    }

    {
      real sum;
      sum  = window[0x0] * b0[0x0];
      sum += window[0x2] * b0[0x2];
      sum += window[0x4] * b0[0x4];
      sum += window[0x6] * b0[0x6];
      sum += window[0x8] * b0[0x8];
      sum += window[0xA] * b0[0xA];
      sum += window[0xC] * b0[0xC];
      sum += window[0xE] * b0[0xE];
      clip += WRITE_SAMPLE_INT32(samples,sum);
      b0-=0x10,window-=0x20,samples+=step;
    }
    window += bo1<<1;

    for (j=15;j;j--,b0-=0x10,window-=0x20,samples+=step)
    {
      real sum;
      sum = -window[-0x1] * b0[0x0];
      sum -= window[-0x2] * b0[0x1];
      sum -= window[-0x3] * b0[0x2];
      sum -= window[-0x4] * b0[0x3];
      sum -= window[-0x5] * b0[0x4];
      sum -= window[-0x6] * b0[0x5];
      sum -= window[-0x7] * b0[0x6];
      sum -= window[-0x8] * b0[0x7];
      sum -= window[-0x9] * b0[0x8];
      sum -= window[-0xA] * b0[0x9];
      sum -= window[-0xB] * b0[0xA];
      sum -= window[-0xC] * b0[0xB];
      sum -= window[-0xD] * b0[0xC];
      sum -= window[-0xE] * b0[0xD];
      sum -= window[-0xF] * b0[0xE];
      sum -= window[-0x0] * b0[0xF];

      clip += WRITE_SAMPLE_INT32(samples,sum);
    }
  }

  (*pnt) += 256;

  return clip;
}

