#include <stdlib.h>
#include <stdio.h>
#include <float.h>

#include "mpg123.h"
#include "mpglib.h"

#include "kbmpglib.h"

#include <windows.h>
#include <mmreg.h>

char*	errorstring = "no error";
int		forceDecoding = 0;

BOOL InitMP3(struct mpstr* mp, int Format, int OutputBps)
{
	static int	InitFormat = -1;
	static int	InitOutputBps = 0;

	memset(mp, 0, sizeof(struct mpstr));

	mp->framesize = 0;
	mp->fsizeold = -1;
	mp->bsize = 0;
	mp->head = mp->tail = NULL;
	mp->fr.single = -1;
	mp->bsnum = 0;
	mp->synth_bo = 1;

	mp->totalFramesize = 0;
	mp->nframe = 0;
	mp->currentBitrate = 0;

	mp->Format = Format;
	mp->OutputBps = OutputBps;

	if((InitFormat != Format) || (InitOutputBps != OutputBps)) {
		long	ScaleVal;

		switch(Format) {
		case DATA_FORMAT_LINEAR_PCM:
			switch(OutputBps) {
			case 16:
				ScaleVal = 0x7fff;
				break;
			case 24:
				ScaleVal = 0x7fffff;
				break;
			case 32:
				ScaleVal = 0x7fffffff;
				break;
			default:
				return 0;
			}
			break;
		case DATA_FORMAT_IEEE_FLOAT:
			switch(OutputBps) {
			case 32:
			case 64:
				ScaleVal = 1;
				break;
			default:
				return 0;
			}
			break;
		}

		make_decode_tables(ScaleVal);
		init_layer3(SBLIMIT);

		InitFormat = Format;
		InitOutputBps = OutputBps;
	}

	_controlfp(_RC_NEAR, _MCW_RC);

	return !0;
}

void ExitMP3(struct mpstr *mp)
{
	struct buf *b,*bn;

	b = mp->tail;
	while(b) {
		free(b->pnt);
		bn = b->next;
		free(b);
		b = bn;
	}
}

static struct buf *addbuf(struct mpstr *mp,char *buf,int size)
{
	struct buf *nbuf;

	nbuf = (struct buf*) malloc( sizeof(struct buf) );
	if(!nbuf) {
		//fprintf(stderr,"Out of memory!\n");
		errorstring = "Out of memory";
		return NULL;
	}
	nbuf->pnt = (unsigned char*) malloc(size);
	if(!nbuf->pnt) {
		free(nbuf);
		return NULL;
	}
	nbuf->size = size;
	memcpy(nbuf->pnt,buf,size);
	nbuf->next = NULL;
	nbuf->prev = mp->head;
	nbuf->pos = 0;

	if(!mp->tail) {
		mp->tail = nbuf;
	}
	else {
	  mp->head->next = nbuf;
	}

	mp->head = nbuf;
	mp->bsize += size;

	return nbuf;
}

static void remove_buf(struct mpstr *mp)
{
  struct buf *buf = mp->tail;
  
  mp->tail = buf->next;
  if(mp->tail)
    mp->tail->prev = NULL;
  else {
    mp->tail = mp->head = NULL;
  }
  
  free(buf->pnt);
  free(buf);

}

static int read_buf_byte(struct mpstr *mp)
{
	unsigned int b;

	int pos;

	pos = mp->tail->pos;
	while(pos >= mp->tail->size) {
		remove_buf(mp);
		pos = mp->tail->pos;
		if(!mp->tail) {
			errorstring = "read_buf_byte error";
			//fprintf(stderr,"Fatal error!\n");
			return 0;
		}
	}

	b = mp->tail->pnt[pos];
	mp->bsize--;
	mp->tail->pos++;
	

	return b;
}

static void read_head(struct mpstr *mp)
{
	unsigned long head;

	head = read_buf_byte(mp);
	head <<= 8;
	head |= read_buf_byte(mp);
	head <<= 8;
	head |= read_buf_byte(mp);
	head <<= 8;
	head |= read_buf_byte(mp);

	mp->header = head;
}

int decodeMP3(struct mpstr *mp,char *in,int isize,char *out,
		int osize,int *done)
{
	int len;

	errorstring = "no error or unknown error";

	if(osize < 4608) {
		errorstring = "output buffer is too small";
		return MP3_ERR;
	}

	if(in) {
		if(addbuf(mp,in,isize) == NULL) {
			return MP3_ERR;
		}
	}


	/* First decode header */
	if(mp->framesize == 0) {
		if(mp->bsize < 4) {
			return MP3_NEED_MORE;
		}
		read_head(mp);
		if (!decode_header(&mp->fr,mp->header)) {
			return MP3_ERR;
		}
		mp->framesize = mp->fr.framesize;
		mp->totalFramesize += mp->framesize+4;
		{
			static int tabsel_123[2][3][16] = {
			   { {0,32,64,96,128,160,192,224,256,288,320,352,384,416,448,},
				 {0,32,48,56, 64, 80, 96,112,128,160,192,224,256,320,384,},
				 {0,32,40,48, 56, 64, 80, 96,112,128,160,192,224,256,320,} },

			   { {0,32,48,56,64,80,96,112,128,144,160,176,192,224,256,},
			     {0,8,16,24,32,40,48,56,64,80,96,112,128,144,160,},
			   {0,8,16,24,32,40,48,56,64,80,96,112,128,144,160,} }
			};

			mp->currentBitrate = tabsel_123[mp->fr.lsf][2][mp->fr.bitrate_index];
		}
		mp->nframe++;
	}

	/*	  printf(" fr.framesize = %i \n",mp->fr.framesize);
		  printf(" bsize        = %i \n",mp->bsize);
	*/

	if(mp->fr.framesize > mp->bsize) {
	  return MP3_NEED_MORE;
	}
	wordpointer = mp->bsspace[mp->bsnum] + 512;
	mp->bsnum = (mp->bsnum + 1) & 0x1;
	bitindex = 0;

	len = 0;
	while(len < mp->framesize) {
		int nlen;
		int blen = mp->tail->size - mp->tail->pos;
		if((mp->framesize - len) <= blen) {
			nlen = mp->framesize-len;
		} else {
			nlen = blen;
		}
		memcpy(wordpointer+len,mp->tail->pnt+mp->tail->pos,nlen);
		len += nlen;
		mp->tail->pos += nlen;
		mp->bsize -= nlen;
		if(mp->tail->pos == mp->tail->size) {
			remove_buf(mp);
		}
	}

	*done = 0;
	if(mp->fr.error_protection) {
       /* skip crc, we are byte aligned here */
      getbits(16);
    }
	if(do_layer3(mp,&mp->fr,(unsigned char *) out,done) < 0 /*== -1*/) {
		return MP3_ERR;
	}

	mp->fsizeold = mp->framesize;
	mp->framesize = 0;

	return MP3_OK;
}

int set_pointer(struct mpstr *mp,/*int ssize,*/long backstep)
{
  unsigned char *bsbufold;
  if(mp->fsizeold < 0 && backstep > 0) {
    fprintf(stderr,"Can't step back %ld!\n",backstep);
    return MP3_ERR; 
  }
  bsbufold = mp->bsspace[mp->bsnum] + 512;
  wordpointer -= backstep;
  if (backstep)
    memcpy(wordpointer,bsbufold+mp->fsizeold-backstep,backstep);
  bitindex = 0;
  return MP3_OK;
}




